;******************* Constants ***********************

MIN_TOP_DISTANCE	EQU	25
MIN_UNDER_WATER		EQU	-10
MAX_Y_CORRECTION	EQU	10
JAGGEDNESS_CORRECTION	EQU	1			;The less the value, the more jagged

;************** ASM file header stuff ***************

	include	"tios.h"
	include	"flib.h"
	xdef	_library
	xdef	tinxlib@0000

;********************************************************
;* Draw a landscape                                     *
;* Input: a5.l = Ptr to field buffer                    *
;*        d6.w = Width of buffer (bytes)                *
;*        d7.w = Height of buffer (pixels)              *
;*        d4.w = Landscape type (0-4, 0=jagged, 4=flat) *
;********************************************************
tinxlib@0000:
generate_landscape:
	movem.l		a0-a6/d0-d7,-(sp)

;* Allocate memory *
	and.l		#$0000FFFF,d6
	lsl.l		#4,d6		;(8 bits per byte, 2 bytes per word => 16 = 2^4)
	move.l		d6,-(sp)
	jsr		tios::HeapAlloc
	addq.l		#4,sp
	move.w		d0,mem_handle
	beq		NoMem
	tios::DEREF	d0,a6

;* Clear the allocated memory *
	move.l		a6,a1
	move.w		d6,d0
	subq.w		#1,d0
\cloop	clr.b		(a1)+
	dbra		d0,\cloop

;* Setup some stuff *
	move.w		d7,d0
	sub.w		#10,d0
	jsr		flib::random
	move.w		d0,(a6)
	move.w		d6,d3
	subq.w		#2,d3		;d3 = right edge
	move.w		d7,d0
	sub.w		#10,d0
	jsr		flib::random
	move.w		d0,(a6)
	move.w		d0,0(a6,d3.w)
	moveq		#0,d5		;d5 = left edge
	
;* We now use recursiveness to fill in the blanks *
	sub.w		#MIN_TOP_DISTANCE,d7
	bsr		recurse
	add.w		#MIN_TOP_DISTANCE,d7

;* Draw the calculated landscape *
	move.w		d6,d3
	lsr.w		#1,d3
	subq.w		#1,d3	;Pixel counter
	lsr.l		#4,d6	;Get back our original width, but in bytes
	move.l		a6,a1	;Pixel pointer
	move.l		a5,a2	;Buffer pointer
	moveq		#7,d0	;Bit to set
\ploop	move.w		(a1)+,d2	;Y
	ble.s		\next		;Check limit #1...
	neg.w		d2
	add.w		d7,d2
	blt.s		\next		;...and #2
	mulu.w		d6,d2
	bset.b		d0,0(a2,d2.w)	;Set pixel
\next	subq.w		#1,d0	;Next bit
	bge.s		\next2
	addq.l		#1,a2	;Next byte
	moveq		#7,d0
\next2	dbra		d3,\ploop

;* Fill in the dirt *
	lea		0(a5,d6.w),a5
	move.w		d6,d5
	lsr.w		#1,d5	;Width in words
	subq.w		#1,d7	;We're skipping one line on the top, don't want it at the bottom!
	mulu.w		d7,d5
	subq.w		#1,d5
	neg.w		d6	;-Width in bytes (to access previous line)
\floop	move.w		0(a5,d6.w),d1
	or.w		(a5),d1
	move.w		d1,(a5)+
	dbra		d5,\floop

;Done - deallocate memory and then quit *
	move.w		mem_handle,-(sp)
	jsr		tios::HeapFree
	addq.l		#2,sp

NoMem	movem.l		(sp)+,a0-a6/d0-d7
	rts

;******************** Subroutines ************************

recurse:
	movem.l		a6/d3-d7,-(sp)

	move.w		d3,d6
	sub.w		d5,d6
	and.w		#%1111111111111100,d6	;Make sure midpoint is an even number
	lsr.w		#1,d6			;Get midpoint's distance from left edge
	beq		\done			;If midpoint = left edge the last recursion has been reached
	move.w		d6,d1			;Store the distance, we'll use it later
	add.w		d5,d6			;Get midpoint's position

	move.w		0(a6,d5.w),d2		;Get left edge's elevation
	add.w		0(a6,d3.w),d2		;Get right edge's elevation
	asr.w		#1,d2			;d2 is now in the middle of the two edges' elevations
	lsl.w		#JAGGEDNESS_CORRECTION,d1
	lsr.w		d4,d1			;Make the landscape just as jagged as we want it to be
	addq.w		#2,d1
	move.w		d1,d0
	jsr		flib::random
	lsr.w		#1,d1
	sub.w		d1,d0			;Get elevation change
	add.w		d0,d2			;Add it to the midpoint's level

	cmp.w		#MIN_UNDER_WATER,d2	;Make sure the sea bottom necer falls under a certain level
	bge.s		\isair
	move.w		#MIN_UNDER_WATER,d2
	bra.s		\isok
\isair	cmp.w		d7,d2			;Make sure we can walk everywhere (d7 = HEIGHT)
	ble.s		\isok			;Elevation is within limits
	move.w		d2,d1			;Elevation is too high, lower it a little
	sub.w		d7,d1
	lsl.w		#1,d1
	sub.w		d1,d2
\isok	move.w		d2,0(a6,d6.w)		;Set midpoint
	exg.w		d6,d3
	bsr		recurse			;Repeat process with left part
	exg.w		d6,d3
	exg.w		d6,d5
	bsr		recurse			;...and with right part
\done	movem.l		(sp)+,a6/d3-d7
	rts

;********************** Variables **************************

	EVEN
mem_handle	dc.w	0
;mem_ptr		dc.l	0

;***************************** Byte stuff *******************************

;************** End of Fargo program ****************

_library:	dc.b	"tinxlib",0

;*****************************************************

	end
